"""beaker-jobwatch - monitor and reschedule beaker jobs."""

import datetime
import functools
import os
import pathlib
import re
import subprocess
import sys
import tempfile
import time
import typing
import xmlrpc.client

import lxml.etree
import requests

from . import cliargs
from . import timeout
from . import utils
from .bkr_wrapper import Bkr

__version__ = "8.0"

STATUS_DOT_COUNT = 30

XMLRPC_CLIENT = xmlrpc.client.ServerProxy("https://beaker.engineering.redhat.com")
SESSION = requests.Session()


def read_options(
    param_options,
    ignore_list,
):
    """Read jobs to run."""
    herd = []
    rs_list = []
    duration = {}

    # populate herd with all recipesets from all given jobs (from both test-run and cmdline)
    for job in set(re.split(r"\W+", ",".join(param_options.jobs))):
        # if input was in TJ#XYZ format, if was split as [TJ,XYZ] - skip the TJ part
        # ditto for J:XYZ and empty jobs
        if job in {"TJ", "J", ""}:
            continue
        print(f"processing job TJ#{job}")
        recipe_sets = RecipeSetCache.add_job(job)
        for rs_id in recipe_sets:
            # check whether was the recipeSet already processed
            if rs_id in rs_list:
                utils.debug(f"already processed recipeSet {rs_id}")
            else:
                utils.debug(f"watching recipeSet {rs_id}")
                rs = RecipeSet(
                    rs_id,
                    job,
                    param_options.max_reschedules,
                    param_options.max_reschedules,
                    param_options.skip_destructive,
                    param_options.skip_machine,
                    ignore_list,
                )
                if rs.is_nacked() and not param_options.load_nacked:
                    utils.debug("recipeSet already NACKed, skipping")
                else:
                    # first recipe for this arch, note the time
                    if rs.description not in duration:
                        duration[rs.description] = {
                            "start": datetime.datetime.now(datetime.UTC),
                            "end": None,
                        }
                    herd.append(rs)
                    # add RS ID to list of already parsed RS
                    rs_list.append(rs_id)
                    utils.debug(f"newly processed recipeSet {rs_id}")

    if len(herd) == 0:
        print("no recipeSets found (all nacked?)")
        sys.exit(1)

    print(f"\nThere was {len(rs_list)} new recipeSet(s) added")

    return len(rs_list), herd, duration


class LogCache:
    """Stores log content from beaker recipes."""

    logs: typing.ClassVar = {}
    log_list: typing.ClassVar = {}

    @classmethod
    def get_log(cls, log_url):
        """Return content of one log identified by a log URL (from cache if available)."""
        # check cache, update if empty
        if log_url not in cls.logs:
            utils.debug(f"downloading log {log_url}")
            response = SESSION.get(log_url)
            response.raise_for_status()
            cls.logs[log_url] = response.content.decode("utf-8")
        return cls.logs[log_url]

    @classmethod
    def get_recipe_logs(cls, recipe_id, log_name="install.log"):
        """Return content of all logs for a particular recipe that matches log_name."""
        # check cache, update if empty
        if recipe_id not in cls.log_list:
            utils.debug(f"getting log list for recipe R:{recipe_id}")
            p = Bkr(["job-logs", f"R:{recipe_id}"]).run_once()
            if p.returncode != 0:
                print(f"retrieving recipe log list failed for R:{recipe_id}")
                return []
            cls.log_list[recipe_id] = list(str(p.stdout).split("\n"))
        log_list = [x for x in cls.log_list[recipe_id] if log_name in x]
        logs = []
        for log_url in log_list:
            try:
                logs.append(cls.get_log(log_url))
            except Exception as e:
                print(f"failed to retrieve log {log_url}: {e}")
        return logs


class RecipeSetCache:
    """Stores recipeset results in a parsed xml as returned by bkr job-results.

    Automatically updates cache for running RSs if older than defined interval.
    """

    cache: typing.ClassVar = {}
    update_interval = 60

    @classmethod
    def add(cls, rs_id):
        """Add a recipeSet to the cache."""
        xml = cls._get_results(rs_id)
        # keep whiteboard if we already have it
        try:
            job_whiteboard = cls.cache[rs_id]["job_whiteboard"]
        except KeyError:
            job_whiteboard = "Unknown"
        try:
            first_seen = cls.cache[rs_id]["first_seen"]
        except KeyError:
            first_seen = datetime.datetime.now(tz=datetime.UTC)
        cls.cache[rs_id] = {
            "xml": xml,
            "time": time.time(),
            "job_whiteboard": job_whiteboard,
            "first_seen": first_seen,
        }

    @classmethod
    def add_job(cls, job_id):
        """Add all recipeSets of a job to the cache.

        Returns list of recipeSet IDs found.
        """
        xml = cls._get_results(job_id, recipe_set=False)
        job_whiteboard = xml.find("whiteboard").text
        rs_ids = []
        for rs in xml.findall(".//recipeSet"):
            rs_id = rs.get("id")
            rs_ids.append(rs_id)
            try:
                first_seen = cls.cache[rs_id]["first_seen"]
            except KeyError:
                first_seen = datetime.datetime.now(tz=datetime.UTC)
            cls.cache[rs_id] = {
                "xml": rs,
                "time": time.time(),
                "job_whiteboard": job_whiteboard,
                "first_seen": first_seen,
            }
        return rs_ids

    @classmethod
    def get(
        cls,
        rs_id,
        *,
        force: bool = False,
        cache_only: bool = False,
    ) -> lxml.etree._Element | None:
        """Return XML for a recipeSet from the cache."""
        if rs_id not in cls.cache:
            return None
        if not force:
            # results are still fresh
            if (time.time() - cls.cache[rs_id]["time"]) < cls.update_interval:
                return cls.cache[rs_id]["xml"]
            if cache_only:
                return cls.cache[rs_id]["xml"]
        # RS is finished, no update necessary
        running = False
        for recipe in cls.cache[rs_id]["xml"].findall(".//recipe"):
            if recipe.get("status") not in ("Completed", "Cancelled", "Aborted"):
                running = True
        if running:
            cls.add(rs_id)
        return cls.cache[rs_id]["xml"]

    # an alias for the get method
    update = get

    @classmethod
    def _get_results(cls, job_or_rs_id, *, recipe_set: bool = True) -> lxml.etree._Element:
        """Get the XML results of a job/recipeSet from Beaker."""
        prefix = "RS"
        if not recipe_set:
            prefix = "J"
        retries = 0
        while True:
            utils.debug(f"fetching xml results of {prefix}:{job_or_rs_id}")
            p = Bkr(["job-results", "--no-logs", f"{prefix}:{job_or_rs_id}"]).run_once()
            if p.returncode == 0:
                break
            print(
                f"retrieving {prefix} results failed for {prefix}:{job_or_rs_id} (try #{retries})",
            )
            retries += 1
            time.sleep(120)
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        return lxml.etree.XML(p.stdout, parser)

    @classmethod
    def first_seen(cls, rs_id):
        """Return timestamp of when recipeSet was added to the cache."""
        return cls.cache[rs_id]["first_seen"]

    @classmethod
    def whiteboard(cls, rs_id):
        """Return job whiteboard for this RS."""
        return cls.cache[rs_id]["job_whiteboard"]

    @classmethod
    def dump_to_disk(cls, rs_id):
        """Dump the XML of a recipeSet to disk."""
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(cls.cache[rs_id]["xml"]))
        file.close()
        print(f"RS {rs_id} dumped to {filename}")


class RecipeSet:
    """Represents a recipeSet in Beaker."""

    def __init__(
        self,
        rs_id,
        job,
        reschedules,
        max_reschedules,
        skip_destructive,
        skip_machine,
        ignore_list=None,
    ):
        """Create a new RecipeSet object."""
        if ignore_list is None:
            ignore_list = []
        self.id = rs_id
        self.job = job
        self.reschedules_left = reschedules
        self.max_reschedules = max_reschedules
        self.status_description = None
        self.active = True
        self.skip_destructive = skip_destructive
        self.skip_machine = skip_machine
        self.description = ""
        self.ignore_list = ignore_list
        xml = self.get_xml()
        for recipe in xml.findall(".//recipe"):
            self.description = f"{self.description},{recipe.get('arch')}"
        self.description = self.description.lstrip(",")
        utils.debug(
            f"new recipeSet created (id={rs_id}, job={job}, "
            f"reschedules_left={reschedules},ignore_list={ignore_list})",
        )
        self.last_ping = {}
        self.last_reboot = {}
        self.reboot_time = 60 * cliargs.OPTIONS.reboot_time
        self.cur_reboots_without_ping = {}

    def __str__(self):
        """Return the recipeSet ID as the string representation."""
        return str(self.id)

    def get_xml(self, *, cache_only: bool = True) -> lxml.etree._Element:
        """Return the parsed XML result of the entire RS."""
        if cliargs.OPTIONS.aggressive:
            cache_only = False
        return RecipeSetCache.get(self.id, cache_only=cache_only)

    def is_ok(self):
        """Return whether no error was found (so far) in this recipeSet.

        Equals to is_finished unless in aggressive mode in which case
        the full results XML is fetched and analyzed.
        """
        # queued for longer than allowed
        if cliargs.OPTIONS.queue_timeout:
            task_info = self.get_task_info()
            if task_info and task_info["state"] == "Queued":
                tracked_for = (
                    datetime.datetime.now(tz=datetime.UTC) - RecipeSetCache.first_seen(self.id)
                ).total_seconds()
                utils.debug(
                    f"RS:{self.id} first seen {RecipeSetCache.first_seen(self.id)}, "
                    f"remaining life: {cliargs.OPTIONS.queue_timeout - tracked_for}",
                )
                if tracked_for > cliargs.OPTIONS.queue_timeout:
                    self.status_description = "queue timeout"
                    return False

        # reboots
        if cliargs.OPTIONS.reboot:
            for box in self.get_machines():
                if subprocess.run(["ping", "-c", "1", box], check=False).returncode == 0:
                    utils.debug(f"Successful ping to {box} ( J:{self.job} / RS:{self.id} )")
                    self.last_ping[box] = time.time()
                    self.cur_reboots_without_ping[box] = 0
                else:
                    if (box in self.last_reboot) and (
                        (time.time() - self.last_reboot[box]) < self.reboot_time
                    ):
                        utils.debug(f"Waiting for {box} to reboot ( J:{self.job} / RS:{self.id} )")
                        continue
                    utils.debug(f"Unsuccessful ping to {box} ( J:{self.job} / RS:{self.id} )")
                # consider rebooting
                if box not in self.last_ping:
                    # here we assume that the box didn't install yet
                    utils.debug(
                        f"Not a single ping to {box} yet ( J:{self.job} / RS:{self.id} ). "
                        f"Skipping it for now.",
                    )
                    continue
                if time.time() - self.last_ping[box] > self.reboot_time:
                    if (
                        self.cur_reboots_without_ping[box]
                        >= cliargs.OPTIONS.max_reboots_without_ping
                    ):
                        print(
                            f"System {box} fails to ping after "
                            f"{self.cur_reboots_without_ping[box]} reboots, rescheduling "
                            f"J:{self.job} / RS:{self.id}.",
                        )
                        # Cause a reschedule() and add the new job to herd (!)
                        self.status_description = f"failed to ping {box} after reboot"
                        return False
                    print(f"Rebooting {box} ( J:{self.job} / RS:{self.id} ) now ...")
                    Bkr(["system-power", "--action", "reboot", box]).run_once()
                    self.last_reboot[box] = time.time()
                    self.cur_reboots_without_ping[box] += 1
                    continue

        # to decrease load on Beaker, don't fetch full xml
        if not cliargs.OPTIONS.aggressive and not self.is_finished():
            return True

        rs_xml = self.get_xml(cache_only=False)

        # destructive task aborting all tasks following it
        if ("destructive-task" not in self.ignore_list) and (
            destructive_tasks := self.get_destructive_tasks()
        ):
            self.status_description = f"destructive task(s) {
                functools.reduce(
                    lambda x, y: f'{x},{y}',
                    [task.get('name') for task in destructive_tasks],
                )
            }"
            return False

        # aborted/paniced recipes
        if "abort" not in self.ignore_list:
            for recipe in rs_xml.findall(".//recipe"):
                if recipe.get("status") == "Aborted":
                    # only check for panics in aborted recipes
                    # panics in completed ones are task-errors but not
                    # job breakers = ok, no need to reschedule
                    if recipe.get("result") == "Panic":
                        self.status_description = "kernel panic"
                    else:
                        self.status_description = "recipe aborted"
                        # check if no HW available for the recipe
                        if (results := recipe.xpath("//result")) and (
                            "does not match any systems" in results[0].text
                        ):
                            self.status_description = "no systems match required HW"
                            # irrecoverable error, don't reschedule
                            self.reschedules_left = 0
                        # check install logs for common errors
                        # check for broken scriptlets
                        logs = LogCache.get_recipe_logs(
                            recipe_id=recipe.get("id"),
                            log_name="install.log",
                        )
                        if len(logs) > 0:
                            for line in logs[0].split("\n"):
                                match = re.match(r"^error:.*\((.*)\) scriptlet failed.*", line)
                                if match:
                                    package_name = match.group(1)
                                    self.status_description = (
                                        f"bad scriptlet in {package_name} package"
                                    )
                                    # irrecoverable error, don't reschedule
                                    self.reschedules_left = 0
                        else:
                            utils.debug(f"no install.log found for recipe R:{recipe.get('id')}")

                        # check for packaging issue (conflicts, missing deps,....)
                        packaging_issues = [
                            {
                                "file": "anaconda.log",
                                "pattern": r".*file conflicts.*installs of ([^ ]*) and ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*packaging: ([^ ]*) conflicts with ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*WARN packaging: ([^ ]*) requires ([^ ]*)",
                                "message": "Package {0} has missing dependency: {1}",
                            },
                        ]
                        for issue in packaging_issues:
                            logs = LogCache.get_recipe_logs(
                                recipe_id=recipe.get("id"),
                                log_name=issue["file"],
                            )
                            if len(logs) > 0:
                                for line in logs[0].split("\n"):
                                    if match := re.match(issue["pattern"], line):
                                        package_name1 = match.group(1)
                                        package_name2 = match.group(2)
                                        self.status_description = issue["message"].format(
                                            package_name1,
                                            package_name2,
                                        )
                                        # packaging issues are irrecoverable, don't reschedule
                                        self.reschedules_left = 0
                            else:
                                utils.debug(
                                    f"no {issue['file']} found for recipe R:{recipe.get('id')}",
                                )
                    # only report as aborted if last task is aborted
                    # otherwise single task in the middle can make recipe aborted
                    # eventhough it's mostly usable [BZ#1327991]
                    last_task = rs_xml.findall(".//task")[-1]
                    if last_task.get("status") == "Aborted":
                        return False

        # failed /distribution/install
        if "install" not in self.ignore_list:
            for task in rs_xml.findall(".//task"):
                if (task.get("name") == "/distribution/install") and (
                    (task.get("result") not in ("Pass", "Warn", "New"))
                    or (task.get("status") in ("Aborted"))
                ):
                    self.status_description = "installation failed"
                    return False

        # critical task failure
        for task in rs_xml.findall(".//task"):
            if task.get("name") in cliargs.OPTIONS.critical_tasks and (
                (
                    task.get("result") not in ("Pass", "New")
                    and task.get("status") not in ("Cancelled")
                )
                or task.get("status") in ("Aborted")
            ):
                self.status_description = f"critical task {task.get('name')} failed"
                # irrecoverable error, don't reschedule
                if task.get("name") in {
                    "/distribution/errata/setup",
                    "/distribution/errata/prepare",
                }:
                    self.reschedules_left = 0
                    self.status_description = (
                        f"critical task {task.get('name')} failed (irrecoverable)"
                    )
                return False

        # external watchdog
        if "watchdog" not in self.ignore_list:
            for task in rs_xml.findall(".//result"):
                if task.text == "External Watchdog Expired":
                    if task.getparent().getparent().get("name") == "/distribution/reservesys":
                        utils.debug(f"ignoring external watchdog in reservesys (RS:{self.id})")

                        continue
                    self.status_description = "killed by external watchdog"
                    return False

        return True

    def get_machines(self):
        """Get list of machines on which this RS ran."""
        return [
            system
            for machine in self.get_xml().findall(".//recipe")
            if (system := machine.get("system")) is not None
        ]

    def get_destructive_tasks(self):
        """Find if there is an aborted task and is followed only by other aborts."""
        rs_xml = self.get_xml()

        first_aborted_task = None
        last_notaborted_task = None
        abort_count = 0
        results = []
        for task in rs_xml.findall(".//task"):
            if task.get("name") not in ("/distribution/install", "/distribution/reservesys"):
                if first_aborted_task is None:
                    if task.get("status") == "Aborted":
                        first_aborted_task = task
                    else:
                        last_notaborted_task = task
                elif task.get("status") != "Aborted":
                    first_aborted_task = None
                    last_notaborted_task = task
                    abort_count = 0
                else:
                    abort_count += 1
        # last_notaborted_task == None => everything is aborted including install => no destrtask
        # first_aborted_task == None => nothing is aborted => no destrtask
        if last_notaborted_task is not None and first_aborted_task is not None:
            # if the task has at least one OK result then it's the culprit. Otherwise
            # it might be its predecessor - the last_notaborted_task
            for result in first_aborted_task.findall(".//result"):
                if not (
                    "LOCALWATCHDOG" in result.get("path")
                    or "Watchdog" in result.text
                    or "Task aborted by task" in result.text
                ):
                    results = [first_aborted_task]
            results = [first_aborted_task, last_notaborted_task]

        # if a destructive task is a setup task then we need to keep it [BZ#1331308]
        new_results = []
        for task in results:
            is_setup = False
            params = task.findall(".//param")
            for param in params:
                if param.get("name") == "WOW_TASK_TYPE" and param.get("value") == "setup":
                    utils.debug(
                        f"not marking {task.get('name')} as destructive because it's a setup task",
                    )
                    is_setup = True
            if not is_setup:
                new_results.append(task)
        return new_results

    def get_task_info(self):
        """Get the task_info object via XMLRPC."""
        tries = 3
        task_info = None
        while tries > 0:
            try:
                # Wait up to 60 seconds for xml client to return
                with timeout.timeout(60):
                    task_info = XMLRPC_CLIENT.taskactions.task_info(f"RS:{self.id}")
                break
            except timeout.AlarmError:
                utils.debug("timeout of get_task_info")
            except Exception:
                time.sleep(10)
                tries -= 1
        return task_info

    def is_finished(self):
        """Return if the recipeSet is no longer running in Beaker."""
        # unless in aggressive mode, don't fetch the whole xml
        if not cliargs.OPTIONS.aggressive:
            task_info = self.get_task_info()
            return bool(task_info and task_info["is_finished"])

        # aggressive mode: heavy-weight check using the full xml result
        finished = True
        rs_xml = self.get_xml()
        # check recipeset status
        for recipe in rs_xml.findall(".//recipe"):
            if recipe.get("status") in {
                "Running",
                "Queued",
                "New",
                "Scheduled",
                "Processed",
                "Waiting",
                "Installing",
            }:
                finished = False
        # check all end-tasks in recipeset
        for task in rs_xml.findall(".//task"):
            if (task.get("name") in cliargs.OPTIONS.end_tasks) and (
                task.get("result") not in {"New"}
            ):
                finished = True
        if finished and self.is_ok():
            self.status_description = "ok"
        return finished

    def is_nacked(self):
        """Return if the recipeSet is nacked in Beaker."""
        rs_xml = self.get_xml()
        return rs_xml.get("response") == "nak"

    def nack(self):
        """Nack the recipeSet in Beaker."""
        utils.debug(f"nacking RS:{self.id}")
        p = Bkr(["job-modify", "-r", "nak", f"RS:{self.id}"]).run_once()
        if p.returncode != 0:
            print(f"nacking RS:{self.id} failed")
            return False
        return True

    def cancel(self):
        """Cancel the recipeSet in Beaker."""
        utils.debug(f"canceling RS:{self.id}")
        p = Bkr(
            [
                "job-cancel",
                f"RS:{self.id}",
                "--msg",
                f"Canceled by beaker-jobwatch: {self.status_description}",
            ],
        ).run_once()
        if p.returncode != 0:
            print(f"canceling RS:{self.id} failed")
            return False
        return True

    def reschedule(self):
        """Reschedule the recipeSet in Beaker."""
        if self.reschedules_left == 0:
            print("recipeSet reached maximum number of reschedules")
            return -1, None, None
        # download vanilla xml job definition
        p = Bkr(["job-clone", "--xml", "--dryrun", f"RS:{self.id}"]).run_repeatedly()
        if p.returncode != 0:
            print(f"getting job definition of RS:{self.id} for cloning failed")
            return -1, None, None
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        job_xml = lxml.etree.XML(p.stdout, parser)

        # update whiteboard
        whiteboard = job_xml.find("whiteboard")
        reason = f"clone of TJ#{self.job}/RS:{self.id}"
        whiteboard.text += f"({reason} #{1 + self.max_reschedules - self.reschedules_left})"

        # set group for the job
        if cliargs.OPTIONS.job_group:
            job_xml.set("group", cliargs.OPTIONS.job_group)
            utils.debug(f"setting job group: {cliargs.OPTIONS.job_group}")

        # set owner for the job
        if cliargs.OPTIONS.job_owner:
            job_xml.set("user", cliargs.OPTIONS.job_owner)
            utils.debug(f"setting job owner: {cliargs.OPTIONS.job_owner}")

        # skip destructive tasks
        if self.skip_destructive:
            for destructive_task in self.get_destructive_tasks():
                for task in job_xml.findall(".//task"):
                    if (task.get("name") == destructive_task.get("name")) and (
                        task.getnext() is not None
                    ):
                        utils.debug(
                            f"removing destructive task {destructive_task.get('name')} "
                            f"(id: {destructive_task.get('id')})",
                        )
                        task.getparent().remove(task)
        # skip broken machine
        if self.skip_machine:
            for elem_hostrequire in job_xml.findall(".//hostRequires"):
                for machine in self.get_machines():
                    elem_and = lxml.etree.SubElement(elem_hostrequire, "and")
                    elem_hostname = lxml.etree.SubElement(elem_and, "hostname")
                    elem_hostname.set("op", "!=")
                    elem_hostname.set("value", machine)
                    utils.debug(f"excluding machine: {machine}")

        # dump to a file for submitting
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(job_xml))
        file.close()

        # submit to beaker as new job
        p = Bkr(["job-submit", filename]).run_repeatedly()
        if p.returncode != 0:
            print(f"failed to submit cloned RS:{self.id}")
            return -1, None
        if not (match := re.match(r"^Submitted: \['J:([0-9]*)'\]", p.stdout)):
            print(f"can't parse job id from job-clone output: {p.stdout}", end=" ")
            print(f"job definition: {filename}, stderr: {p.stderr}")
            return -1, None, None
        pathlib.Path(filename).unlink()
        job = match.group(1)
        utils.debug(f"new job TJ#{job} scheduled")
        recipe_set = RecipeSetCache.add_job(job)[0]
        if not self.is_finished():
            self.cancel()
        self.nack()
        self.active = False
        return int(job), reason, recipe_set

    def task_statuses(self):
        """How many tasks are in particular status."""
        statuses = {
            "New": 0,
            "Processed": 0,
            "Queued": 0,
            "Scheduled": 0,
            "Waiting": 0,
            "Running": 0,
            "Completed": 0,
            "Cancelled": 0,
            "Aborted": 0,
        }
        for task in self.get_xml().findall(".//task"):
            statuses[task.get("status")] += 1
        return statuses

    def recipe_ids(self):
        """Return list of all recipes that are part of this RS."""
        return [recipe.get("id") for recipe in self.get_xml().findall(".//recipe")]

    def id_string(self) -> str:
        """Return a string representing this RS's ID.

        For a RS with a single recipe, return the recipe ID instead
        because it's more easily converted to URL (via doer).
        """
        if len(self.recipe_ids()) == 1:
            return f"R:{self.recipe_ids()[0]}"
        return f"RS:{self.id}"


def main(args: list[str] | None = None) -> None:
    """Run the beaker-jobwatch script."""
    print(f"beaker-jobwatch v{__version__}")

    cliargs.parse_args(args)

    # to evaluate end tasks, we need to fetch tasks results repeatedly
    # so turn aggressive mode on
    if len(cliargs.OPTIONS.end_tasks) > 1:
        cliargs.OPTIONS.aggressive = True
    utils.debug(f"options: {cliargs.OPTIONS}")

    if cliargs.OPTIONS.reboot:
        utils.debug(f"reboot time: {cliargs.OPTIONS.reboot_time} min")

    if "BEAKER_USERNAME" in os.environ:
        Bkr.username = os.environ["BEAKER_USERNAME"]
    if "BEAKER_PASSWORD" in os.environ:
        Bkr.password = os.environ["BEAKER_PASSWORD"]
    if cliargs.OPTIONS.username:
        Bkr.username = cliargs.OPTIONS.username
    if cliargs.OPTIONS.password:
        Bkr.password = cliargs.OPTIONS.password

    ignore_list = []
    if cliargs.OPTIONS.ignore_list:
        for option in ",".join(cliargs.OPTIONS.ignore_list).split(","):
            if option not in ("install", "abort", "watchdog", "destructive-task", "none"):
                print(
                    f"Invalid ignore-failure value {option}. Valid values are: "
                    f"install,abort,watchdog,destructive-task",
                )
                sys.exit(1)
        ignore_list = ",".join(cliargs.OPTIONS.ignore_list).split(",")
        if "none" in ignore_list:
            ignore_list = []

    RecipeSetCache.update_interval = cliargs.OPTIONS.sleep - 1
    # list of successfully finished recipesets - will constitute results
    herd_finished_ok = []
    # list of unsuccessfully finished recipes
    herd_broken = []

    # dots are printed as a sort of status if nothing changes
    # lets count how many dots have we printed and add line break after 30
    dot_counter = 0

    # print status even if nothing changed since the last iteration, set to
    # True to print status at start
    force_status = True

    # good_recipesets: how many OK recipesets we must have to finish successfully
    # herd: list of recipesets we monitor - not finished yet
    # duration: measure how long does each arch take to finish
    (good_recipesets, herd, duration) = read_options(cliargs.OPTIONS, ignore_list)

    # main cycle iterating over all recipesets in herd until all are done =
    # = moved to other herds
    while True:
        fingerprint = str([herd, herd_finished_ok, herd_broken])

        # finished successfully
        new_herd = []
        for rs in herd:
            if rs.is_finished() and rs.is_ok():
                if cliargs.OPTIONS.debug:
                    RecipeSetCache.dump_to_disk(rs.id)
                herd_finished_ok.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # broken (running or finished, doesn't matter)
        new_herd = []
        for rs in herd:
            if not rs.is_ok():
                if rs.reschedules_left > 0:
                    if not cliargs.OPTIONS.dry_run:
                        new_job, _reason, rs_id = rs.reschedule()
                        if new_job > 0:
                            utils.debug(f"watching new recipeSet {rs_id}")
                            herd.append(
                                RecipeSet(
                                    rs_id,
                                    new_job,
                                    rs.reschedules_left - 1,
                                    cliargs.OPTIONS.max_reschedules,
                                    rs.skip_destructive,
                                    rs.skip_machine,
                                    ignore_list,
                                ),
                            )
                # normally cancellation happens after successful reschedule,
                # here we do it for recipes that won't be rescheduled (e.g. queue timeouts)
                elif not rs.is_finished():
                    rs.cancel()
                    rs.nack()
                    rs.active = False
                if cliargs.OPTIONS.debug:
                    RecipeSetCache.dump_to_disk(rs.id)
                herd_broken.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # print status
        status = ""
        new_fingerprint = str([herd, herd_finished_ok, herd_broken])
        if force_status or fingerprint != new_fingerprint:
            force_status = False
            # header
            print("\n")
            status += f"{datetime.datetime.now(tz=datetime.UTC).strftime('%Y-%m-%d %H:%M:%S')}\n"
            whiteboards = []
            for rs in herd + herd_finished_ok + herd_broken:
                whiteboard = RecipeSetCache.whiteboard(rs.id)
                # strip whiteboard of the postfix we add during rescheduling ("clone of ...")
                whiteboard = re.sub(r" *\(clone of.*\) *", "", whiteboard)
                if whiteboard not in whiteboards:
                    status += f"{whiteboard}\n"
                    whiteboards.append(whiteboard)
            # individual herds
            status += f"Broken: {len(herd_broken)}\n"
            for rs in herd_broken:
                attempts_string = f"{cliargs.OPTIONS.max_reschedules - rs.reschedules_left}"
                # otherwise hard to spot from log if default (=3) wasn't used
                if rs.reschedules_left == 0:
                    attempts_string += " (max)"
                status += (
                    f"\tTJ#{rs.job} {rs.id_string()} ({rs.description}): "
                    f"broken #{attempts_string}: {rs.status_description}\n"
                )

            status += f"Running:   {len(herd)}/{good_recipesets}\n"
            for rs in herd:
                stats = rs.task_statuses()
                tasks_finished = stats["Aborted"] + stats["Cancelled"] + stats["Completed"]
                tasks_all = functools.reduce(lambda x, y: x + y, list(stats.values()))
                # unless in aggressive mode we don't have up2date task info so ignore it
                if cliargs.OPTIONS.aggressive:
                    tasks_stat_string = f"({tasks_finished}/{tasks_all})"
                else:
                    tasks_stat_string = f"({tasks_all})"
                status += (
                    f"\tTJ#{rs.job} {rs.id_string()} ({rs.description}): "
                    f"running {tasks_stat_string}\n"
                )

            status += f"Completed: {good_recipesets - len(herd)}/{good_recipesets}\n"
            if len(herd) + len(herd_finished_ok) == 0:
                print("no jobs in herd + herd_finished_ok, this should never happen")
                print("(unless you inspect finished broken jobs only)")
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd)])
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd_finished_ok)])
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd_broken)])
            else:
                if len(herd_broken) + len(herd_finished_ok) != 0:
                    status += f"\t{
                        functools.reduce(
                            lambda x, y: f'{x} {y}',
                            {f'TJ#{rs.job}' for rs in (herd_broken + herd_finished_ok)},
                        )
                    }\n"
                status += (
                    f"https://beaker.engineering.redhat.com/matrix/?"
                    f"toggle_nacks_on=on&job_ids={
                        functools.reduce(
                            lambda x, y: f'{x}+{y}',
                            {rs.job for rs in (herd + herd_finished_ok + herd_broken)},
                        )
                    }"
                )
            print(status)

            # check if some arch isn't completely finished, note the time
            for arch, arch_duration in duration.items():
                if arch not in [x.description for x in herd] and arch_duration["end"] is None:
                    arch_duration["end"] = datetime.datetime.now(tz=datetime.UTC)
        else:
            sys.stdout.write(".")
            dot_counter += 1
            if dot_counter == STATUS_DOT_COUNT:
                print(f"\n{datetime.datetime.now(tz=datetime.UTC).strftime('%Y-%m-%d %H:%M:%S')}")
                dot_counter = 0
        sys.stdout.flush()

        if not herd:
            break

        time.sleep(cliargs.OPTIONS.sleep)

    # dump herds for parsing out the (un)successful recipe sets
    utils.debug(f"herd_broken: {','.join(str(x) for x in herd_broken)}")
    utils.debug(f"herd_finished_ok: {','.join(str(x) for x in herd_finished_ok)}")
    # print duration
    delta = max(d["end"] for d in duration.values()) - min(
        d["start"] for d in duration.values()
    )
    print(f"duration: {delta}")
    for arch, arch_duration in duration.items():
        utils.debug(f"duration of {arch}: {arch_duration['end'] - arch_duration['start']}")

    if herd_broken:
        print("finished unsuccessfully")
        sys.exit(2)
    print("finished successfully")
