"""Argument parsing for beaker-jobwatch."""

import argparse
import dataclasses


@dataclasses.dataclass
class Options:
    """CLI arguments."""

    jobs: list[str]
    max_reschedules: int
    dry_run: bool
    debug: bool
    no_wait: bool
    sleep: int
    skip_destructive: bool
    skip_machine: bool
    load_nacked: bool
    ignore_list: list[str]
    reboot: bool
    reboot_time: int
    max_reboots_without_ping: int
    critical_tasks: list[str]
    end_tasks: list[str]
    username: str | None
    password: str | None
    job_group: str | None
    job_owner: str | None
    aggressive: bool
    queue_timeout: int | None


def parse_args(args: list[str] | None = None) -> Options:
    """Parse CLI args and populate OPTIONS."""
    usage = """beaker-jobwatch watches beaker jobs and reschedules any recipe-sets that fail because
    of test-environment problems.

    beaker-jobwatch --job 127346[,TJ#1267,3575,...]
    """
    parser = argparse.ArgumentParser(description=usage)
    parser.add_argument(
        "-j",
        "--job",
        dest="jobs",
        action="append",
        required=True,
        help="specify job(s) to be investigated",
    )
    parser.add_argument(
        "-m",
        "--max",
        dest="max_reschedules",
        metavar="MAX",
        type=int,
        default=3,
        help="Maximum number of times one recipe set will be rescheduled",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Do not reschedule, cancel or nack failed recipesets, do not update run notes",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Very verbose logging",
    )
    parser.add_argument(
        "--sleep",
        type=int,
        default=600,
        help="""How long to wait before polling beaker for updated results (default: 600 seconds),\n
    pressing ENTER will skip the pause and force an update""",
    )
    parser.add_argument(
        "--skip-destructive-task",
        dest="skip_destructive",
        action="store_true",
        help="Don't reschedule task(s) suspected of breaking the recipe",
    )
    parser.add_argument(
        "--skip-broken-machines",
        dest="skip_machine",
        action="store_true",
        help="Instruct Beaker to not reuse machine(s) that ran the failed recipe set.",
    )
    parser.add_argument(
        "--load-nacked",
        action="store_true",
        help="Don't ignore nacked recipesets in specified jobs",
    )
    parser.add_argument(
        "--ignore-failure",
        dest="ignore_list",
        action="append",
        default=["install"],
        help="Comma separated list of checks that will be ignored, these will not cause jobs to "
        "be canceled and rescheduled. Can be any combination of install, abort, watchdog and "
        "destructive-task or none. Default: install.",
    )
    parser.add_argument(
        "--reboot",
        action="store_true",
        help="Reboot unresponsive machines",
    )
    parser.add_argument(
        "--reboot-time",
        type=int,
        default=60,
        help="Reboot time treshold (defaults to 60 minutes)",
    )
    parser.add_argument(
        "--max-reboots-without-ping",
        type=int,
        default=3,
        help="Reboot count treshold for machines not giving a ping",
    )
    parser.add_argument(
        "--critical-task",
        dest="critical_tasks",
        action="append",
        default=[""],
        help="Failure of this task will trigger recipe cancellation, nacking and reschedule",
    )
    parser.add_argument(
        "--end-task",
        dest="end_tasks",
        action="append",
        default=[""],
        help="Consider recipeset finished if an end task has completed, don't wait for the rest of "
        "the recipeset. Requires aggressive mode.",
    )
    parser.add_argument(
        "--username",
        help="Username for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_argument(
        "--password",
        help="Password for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_argument(
        "--job-group",
        help="Reschedule jobs as group jobs belonging to given group",
    )
    parser.add_argument(
        "--job-owner",
        help="Reschedule jobs as delegated to the given owner",
    )
    parser.add_argument(
        "--aggressive",
        action="store_true",
        help="Fetch full results repeatedly, not just after recipe finishes. Stresses Beaker a "
        "lot, use with caution!",
    )
    parser.add_argument(
        "--queue-timeout",
        type=int,
        help="Abort recipe set when queued for longer than defined (in seconds). Default: not set",
    )
    global OPTIONS  # noqa: PLW0603 pylint: disable=global-statement
    OPTIONS = Options(**vars(parser.parse_args(args)))
    return OPTIONS


OPTIONS: Options
